#####################################################################
#:CQCC::MsgCat message catalog management
#
# Licensed Materials - Property of IBM
# Rational ClearCase
# (C) Copyright IBM Corp. 2001, 2008. All Rights Reserved.
# US Government Users Restricted Rights
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#

=head1 NAME

CQCC::MsgCat - Abstract base class to manage all user interaction
communications message text used by IOBase methods Message() and 
Prompt(). 

=head1 CLASS DESCRIPTION

This class provides an abstract base class providing management
methods for all message catalog subclasses making it easier for future
translations and reviewing user messages for consistency.  Each
subclass implementing this class must provide an Initialize() method
that loads the catalog.

The primary message catalog is the MsgCat_English subclass which
provides the messages required for the supported integration.
Subclasses providing translations of MsgCat_English may be used
instead as long as they provide all the message tags needed by the
application. Extension subclasses of MsgCat_English may be used to
provide additional messages beyond those needed by the basic
integration. NOTE: This class does NOT support unicode string handling
so only languages using the standard English/Latin character set are
supported.

In order to be used, the configuration file must ask for a subclass
of this class as shown below:

    &SetConfigParm("MsgCatInterface", "CQCC::MsgCat_English");

SUPPORT POLICY: See TriggerCQCC.pm for the Rational Support Policy. 

=head1 METHODS

=cut

######################################################################
#::Package declarations
######################################################################
package CQCC::MsgCat;
use Exporter;
@ISA = ('Exporter');
@EXPORT = qw(&MsgCat_Prompts);

use strict;

use CQCC::Parm;       # Implicitly included in iobase
use CQCC::IOBase;
sub Version { return '2.2' };

require Config;

######################################################################
#::Class Globals 
######################################################################

my $perl5008006 = ($] ge "5.008006" ? 1 : 0);
my $iobase_ui = "";      # Type of interaction - text or GUI
my $yes_prompt = "";     # Name of keyboard key or button for yes
my $no_prompt = "";      # Name of keyboard key or button for no 

my $CQCCext_available;   # If undefined, check to load CQCCext
                         # Set to 1 is available, 0 if not
                         # If not available, then no conversions attempted.

my $convertCounter = 0;
my $convertCounterUTF8 = 0;  # Number of times convert to utf8 called
my $convertCounterLocal = 0;  # Number of times convert to local called

######################################################################
#::function ClassInit()

=over 6

=item C<ClassInit()>

This class function is automatically invoked to define class 
parameters using CQCC::Parm objects. Current parameters include: 

=over 4

=item C<MsgCatInterface>

Name of MsgCat subclass to use for this session, used by the 
IOClassFactory(). Set to 'CQCC::MsgCat_English' by default.

=back

=back

=cut

######################################################################

sub ClassInit { 
    &SetParmDefaults(PCLASS=>"CQCC::MsgCat");

    DefineParm(NAME=>"MessageCatalogInterface",
               DEFAULT=>"CQCC::MsgCat_English",
               SRC_CFG=>1, SRC_EV=>1);

    DefineParm(NAME=>"CLEARCASE_ENABLE_ATE_MESSAGES",
               DEFAULT=>"FALSE", RANGE_LIST => ["FALSE", "", "TRUE", "True", "true"], 
               SRC_EV=>1);

    &SetParmDefaults();
}

&ClassInit();

######################################################################
#::method Encoding

=over 6

=item C<codepage = Encoding(type, {codepage} )>

This method sets or returns information about data encoding. 
The current values are supported: 

"IO_CODEPAGE", codepage  - defines default output codepage to be returned
by Message method when accessing the message catalog. If codepage is 
not provided, then the current codepage value is returned. 

The codepage may be one of the following:

"CORE_CODEPAGE" (use default codepage used by main body of code). 
NOTE: Initial implementation assumes local codepage but subject to change.

"LOCAL_CODEPAGE" (use local codepage)
NOTE: Initial implementation requires that CQCC_CODEPAGE be set.
NOTE: !!! Need to better distinquish "MY codepage" from "LOCAL codepage"

"UTF8_CODEPAGE" (use UTF8 encoding)

=back

=cut

######################################################################

sub Encoding {
    my ($self, $type, $codepage) = @_;
    if ($type eq "IO_CODEPAGE") {
        if ($codepage ne "") {
            if ($codepage eq "CORE_CODEPAGE" || $codepage eq "LOCAL_CODEPAGE" ||  
                $codepage eq "UTF8_CODEPAGE") {
                $self->{IO_CODEPAGE} = $codepage;
            } else {
                # Note: Use ParmIOMessage since we can't include IOBase here
                &CQCC::Parm::ParmIOMessage("ERROR", ["MC_E_UNKNOWN_ENCODING", 
                    $type, $codepage]);
            }
        }
        return $self->{IO_CODEPAGE};
    }else {
        # Note: Use ParmIOMessage since we can't include IOBase here
        &CQCC::Parm::ParmIOMessage("ERROR", ["MC_E_UNKNOWN_ENCODING", 
            $type, $codepage]);
        return undef;
    }
}

######################################################################
#::method ConvertString

=item C<ConvertString($src_string_or_array, $src_encoding, $dst_encoding)>

Takes a string or string array reference in the given src_encoding and 
converts it to the dst_encoding. 

These are the currently defined encodings: 

"IO_CODEPAGE" - the codepage that the IO class expects, to be determined
                at run time.

"LOCAL_CODEPAGE" - local codepage reference

"UTF8_CODEPAGE" - UTF8 codepage

"CORE_CODEPAGE" - Internal standard codepage

See MsgCat::Encoding for more details. 

=back

=cut

# Internal notes: 
# (1) The MsgCat keeps track of the CORE_CODEPAGE which is the basic
#     assumption about the internal data representation shared in the core.
#     Changing this requires a review for how all modules encode data
# (2) All modules should have a notion of what their own codepage is
#     with CORE assumed by default. Each module should handle converting
#     its values to CORE. 
######################################################################

sub ConvertString {
    my ($self, $src_input, $src_encoding, $dst_encoding) = @_;

    # Determine if input is a scalar or array of strings

    my @strings;
    my @results;
    my $return_type = "scalar";
    my $status; 

    $convertCounter++;

    if (ref($src_input) eq "ARRAY") { 
        @strings = @{ $src_input };
        $return_type = "array";
    } else {
        push @strings, $src_input;
    }

    ##################################################
    # If conversion not enabled, return immediately
    if ($self->{CONVERSION_ENABLE} ne 1) {
        $CQCCext_available = 0;
    }

    # If CQCCext is not available, no conversions possible
    if (! defined $CQCCext_available) { 
        eval { 
             require CQCCext;
             $CQCCext_available = 1;
        };
        if ($@) {
             $CQCCext_available = 0;
        }

        #TEMP: For development debugging only
        if (&GetParm("CQCC_DEBUG") > 2) {
            print "ConvertString: CQCCext.pm " . 
                  ($CQCCext_available ? "IS" : "is NOT") . 
                  " available for to support conversions";
        }
    }     

    if ($CQCCext_available ne 1) {
        if ($self->{CONVERSION_ASCII_CHECK} eq 1) {
            my @newStrings;
            for (@strings) {
                my ($substatus,$result) = 
                    $self->confirmASCII($src_encoding,$_);
                if ($substatus) {
                    $status .= $substatus;                    
                }
                push @newStrings,$result;
            }
            return ($self->truncateStatus($status,$dst_encoding),
                    @newStrings);
        }
        return (undef,@strings);
    }

    ##################################################
    # Convert if needed 
    # If Encode package not available, then don't do translation
    if ($src_encoding eq "IO_CODEPAGE") {
        $src_encoding = $self->{IO_CODEPAGE};
    }
    if ($dst_encoding eq "IO_CODEPAGE") {
        $dst_encoding = $self->{IO_CODEPAGE};
    }

    if ($src_encoding eq "CORE_CODEPAGE") {
        $src_encoding = $self->{CORE_CODEPAGE};
    }
    if ($dst_encoding eq "CORE_CODEPAGE") {
        $dst_encoding = $self->{CORE_CODEPAGE};
    }

    if (! (($src_encoding eq "LOCAL_CODEPAGE" || $src_encoding eq "UTF8_CODEPAGE") &&
           ($dst_encoding eq "LOCAL_CODEPAGE" || $dst_encoding eq "UTF8_CODEPAGE"))) {
        # Note: Use ParmIOMessage since we can't include IOBase here
        &CQCC::Parm::ParmIOMessage("FATAL",["MC_E_WRONG_CODEPAGE",
            $src_encoding, $dst_encoding]);
        # FATAL should exit without returning
        return $status, "";
    }

    if ( $src_encoding eq $dst_encoding || !$perl5008006 ) {
        @results = @strings;                
    } else {
        # the encoding converts the string(s) in either case
        # local codepage => utf8  <or> utf8 => local codepage    

        # Accumulates list of status errors
        # then these are consolidated into one string for return.

        my (@statusList, $substatus,$result);
        my ($str, $substr);
        for $str (@strings) {
            if (ref($str) eq "ARRAY") { 
                my @subresults;
                my @substrings = @{ $str };
                for $substr (@substrings) { 
                    ($substatus, $result) = 
                        $self->_doConvert($dst_encoding, $substr);
                    push @subresults, $result;
                    push @statusList, $substatus if ($substatus);
                }
                push @results, \@subresults;
            } else {
                ($substatus, $result) = 
                    $self->_doConvert($dst_encoding, $str);
                push @results, $result;
                push @statusList, $substatus if ($substatus);
            }
        }
        # If any status items then combine into one string
        # and expect that results have "safe" tokens instead. 
        if (@statusList) {
            my $statusString = join("\n", @statusList);

            # Return status preconverted for inclusion into error reporting
            $status = $self->Message(["MC_E_CONVERSION_ERROR", 
                $src_encoding, $dst_encoding,
                $self->truncateStatus($statusString,$dst_encoding)]);
        }
    }

    # Return array if provided an array, else scalar
    if ($return_type eq "array") {
        return $status, @results;
    } else {
        return $status, $results[0];
    }
}

######################################################################
# EnableConversion: Set by IOBase to turn on conversion

sub EnableConversion {
    my ($self,$i18n_flag,$ascii_flag,$error_max) = @_;
    $self->{CONVERSION_ENABLE} = $i18n_flag;
    $self->{CONVERSION_ASCII_CHECK} = $ascii_flag;
    $self->{CONVERSION_ERROR_MAX} = $error_max;
}

######################################################################
# GetCounters - internal routine to retrieve conversion counts

sub GetCounters {
    return "Counter: $convertCounter ToUTF8: $convertCounterUTF8 " .
           "ToLocal: $convertCounterLocal";
}

######################################################################
# TruncateStatus 

sub truncateStatus {
    my ($self,$status,$encoding) = @_;
    my $maxLines = $self->{CONVERSION_ERROR_MAX};
    if ($status && $maxLines =~ /^\d+$/) {
        my @lines = split("\n",$status);
        if (($#lines+1) > $maxLines) {
            my @newLines;
            my $i;
            for ($i = 0; $i < $maxLines; $i++) {
                push @newLines, $lines[$i];
            }
            my $tail = $self->Message(["MC_E_TRUNCATION_MARKER"],
                                      $encoding);
            push @newLines,$tail;
            $status = join("\n",@newLines);
        }
    }
    return $status;    
}

######################################################################
#::method _doConvert
# Internal method that should only be called by ConvertString
# 
# Inputs: 
#    encoding - target encoding
#    src - string to be converted
#
# Outputs:
#    status - error message if conversion failed, else undef
#    result - converted string if conversion succeeded,
#             else ASCII-safe conversion failed token

sub _doConvert {
    my ($self, $encoding, $inSrc) = @_;
    my ($status,$result);
    my $done = 0;
    my $lastErrorCode = "";
    my $lastLocation = -1;
    my @errorCodes;
    my @complaints;
    my $src = $inSrc;

    while (! $done) { 
        eval { 
            require CQCCext;
            require Encode;

            # If needed, 
            #    do conversion using CQCCext conversion and mark appropriately
            #    else return original values 

            # Convert to UTF8 codepage if needed
            if ($encoding eq "UTF8_CODEPAGE") {
                $convertCounterUTF8++;
                if (Encode::is_utf8($src)) {
                    $result = $src;
                } else {
                    $result = CQCCext::convert_local_to_utf8($src);
                    Encode::_utf8_on($result);
                }
            # Convert to local codepage if needed
            } else {    
                $convertCounterLocal++;
                if (Encode::is_utf8($src)) {
                    $result = CQCCext::convert_utf8_to_local($src);
                    Encode::_utf8_off($result);
                } else {
                    $result = $src;
                }
            }
        };
        if ($@) {
            my ($errorCode,$location);
            if ($@ =~ /Error: (\S*) Location: (\S*)/) {
                my $diffErrorCode = 0;
                $errorCode = $1;
                $location = $2 + 1; # Errors are 1-based indexing

                # Only handle errors with the same error code
                if ($lastErrorCode eq "") {
                    $lastErrorCode = $errorCode;
                    push @errorCodes, $errorCode;
                } elsif ($lastErrorCode ne $errorCode) {
                    $diffErrorCode = 1;
                }
                if (($diffErrorCode ne 0) || 
                    ($errorCode ne "I18N_ICONVEX_INPUT_CONV_ERROR" &&
                     $errorCode ne "I18N_ICONVEX_OUTPUT_CONV_ERROR")) {
                    $src = $self->Message(["MC_E_CONVERSION_MARKER"],
                                          $encoding);
                    $done = 1;
                }
                # If error is one that provides a location, 
                # note the problem, replace the errant char and retry
                # to see how many chars are a problem. 

                if (! $done) {
                    # Avoid getting stuck on same location
                    if ($location eq $lastLocation) {
                        $src = $self->Message(["MC_E_CONVERSION_MARKER"],
                                              $encoding);
                        $done = 1;
                    } else {
                        my @charArray = split(//,$src);
                        my $badChar = $charArray[$location];                    
                        my @unpackArray = unpack("C*",$badChar);
                        my $errChar;
                        for (@unpackArray) {
                            my $e = sprintf "%x",$_;
                            $errChar .= $e;
                        }
                        #TODO: Use msgcat to format "value AT location"
                        push @complaints, "U+$errChar at $location";
                        $charArray[$location] = "?";
                        $src = join("",@charArray);
                        $lastLocation = $location;
                    }
                }
            } else {   # else couldn't parse the error message
                $lastErrorCode = "I18N_ICONVEX_UNKNOWN_ERROR";
                push @complaints, $@;
                $src = $self->Message(["MC_E_CONVERSION_MARKER"],
                                      $encoding);
                $done = 1;
            }
        } else { # else no error
            $done = 1;
        }
    }
    if ($lastErrorCode) {  
        # Return a status message preconverted for inclusion into error reporting
        my $errorMessage = "MC_E_$lastErrorCode";
        my $errors = join("\n",@complaints);
        $status = $self->Message([$errorMessage, $src, $errors ]);
        $result = $src;
    }
    return ($status,$result);
}

######################################################################
#::function confirmASCII
# Checks chars in string to make sure they are ASCII only
# and if not marks them with ?'s and tries remaining characters.
# If UTF8 source string, will replace each multibyte char with one ?
# With local code page will mark initial char with >127 byte as ?
# and may show garbage for following bytes. 

sub confirmASCII {
    my ($self,$src_encoding, $str) = @_;

    my @unpackArray = unpack("C*",$str);
    my @fixedString;
    my @complaints;
    my $unpackLength = $#unpackArray+1;
    my ($status,$result,$i);
    my $charCount = 0;

    for ($i = 0; $i < $unpackLength; $i++) {
        my $charCode = $unpackArray[$i];
        if ($charCode > 127) {
            push @fixedString,ord("?");
            my $loc = $i;
            my $bytesLeft;
            if ($src_encoding eq "UTF8_CODEPAGE") {
                $bytesLeft = &multiByteCount($charCode);
                # Output location is relative to the output string
                $loc = $charCount;
            } else {
                $bytesLeft = 0;
            }
            my $errChar .= sprintf "%x",$charCode;
            while ($bytesLeft > 0) {
                $i++;
                $bytesLeft--;
                if ($i < $unpackLength) {
                    $charCode = $unpackArray[$i];
                    $errChar .= sprintf "%x",$charCode;
                } else {
                    $charCode = 0;
                }
            }
            #TODO: Use msgcat to format "value AT location"
            push @complaints, "U+$errChar at $loc";
        } else {
            push @fixedString,$charCode;
        }
        $charCount++;
    }
    if (@complaints) {
        $result = pack("C*",@fixedString);
        # Return a status message preconverted for inclusion into error reporting
        my $errorMessage = "MC_E_NON_ASCII_I18N_DISABLED_" . 
                           ($perl5008006 ? "CC7" : "PRE_CC7");
        my $errors = join("\n",@complaints);
        $status = $self->Message([$errorMessage, $result, $errors ]);
        return ($status,$result);
    }
    # Else return empty status and unmodified string
    return (undef,$str);
}

######################################################################
#::function multiByteCount 
# Multi-byte chars use convention that high bit (7) is marked
# and number of bits after 7 before 0 indicates remaining bytes
# so 11100000 would indicate two more bytes, etc. 
# 
# If at least one more char (bit 7 and 6 are not 0), keep going

sub multiByteCount {
    my ($charCode) = @_;
    my $result = 0;
    if ($charCode & (1 << 7)) {
        if ($charCode & (1 << 6)) {
            $result++;
            if ($charCode & (1 << 5)) {
                $result++;
                if ($charCode & (1 << 4)) {
                    $result++;
                    if ($charCode & (1 << 3)) {
                        $result++;
                        if ($charCode & (1 << 2)) {
                            $result++;
                            if ($charCode & (1 << 1)) {
                                $result++;
                            }
                        }
                    }
                }
            }
        }
    }
    return $result; 
}

######################################################################
#::method Initialize
######################################################################

sub Initialize {
} 

######################################################################
#::method new

=over 6

=item C<new CQCC::MsgCat()>

This method initializes the objects attributes needed to store
the subclasses messages (CATALOG), and arguments (ARGLISTS). 

=back

=cut

######################################################################

sub new { 
    my $proto = shift;
    my $class = ref($proto) || $proto;  # Object or class name

    my %catalog = ();
    my %arglists = ();

    # Attributes
    my $self = {
        CLASSNAME => "CQCC::MsgCat", 
        CATALOG => \%catalog,   # Hash list holding message entries
        ARGLISTS => \%arglists, # Hash list holding argument lists
        SELFTEST => 0,          # Internal flag to enable self testing

        # Codepage management
        CORE_CODEPAGE => "LOCAL_CODEPAGE", # CODEPAGE if local codepage, else UTF8
        MSGCAT_CODEPAGE => "LOCAL_CODEPAGE", # Assume catalog is in local codepage
        CONVERSION_ENABLE => undef, # Conversion isn't supported unless this is 1
        CONVERSION_ASCII_CHECK => undef, # Confirm strings are ASCII if conversion off
        CONVERSION_ERROR_MAX => undef,  # Max lines in conversion errors 
        @_,
        @_,                     # Extra time for ccperl
    };
      
    bless($self, $class);
    return $self;
}

######################################################################
#::function MsgCat_Prompts

=over 6

=item C<MsgCat_Prompts($type,$yes,$no)>

This function is used by IOBase classes to influence the prompt
strings that tell the user how to select yes (hit YES or hit Enter)
and no.  It must be called before the Message Catalog object is
created so that these values can be burned into the message catalog.

Class should be the class such as "text" or "GUI".

$yes and $no should be the name of the keyboard key or button name. 

=back 

=cut 

######################################################################

sub MsgCat_Prompts { 
    # Only set the value the first time, since this only affects
    # these strings before initialization. 

    if (@_ && $iobase_ui eq "") { 
       ($iobase_ui, $yes_prompt, $no_prompt) = @_;
    } 
    return ($iobase_ui, $yes_prompt, $no_prompt);
}

######################################################################
#::method Message

=over 6

=item C<Message(msgtag, {encoding})>

This method translates message catalog "msgtag" values into translated
and converted strings. A message tag is an array reference consisting
of a message id and a set of string arguments to be combined with the 
message catalog entry associated with that id. A message tag can also 
be a scalar string which is not part of the message catalog but which
will be returned after appropriate conversion.

This method will look up the given msgtag in the message catalog,
format its arguments into it, and return the given string, converting
the value to the output code page if required. 

If the message is not found, an error message will be returned in its
place.  If the number of arguments provided is insufficient, an error
message will also be generated.

If the optional encoding is provided that code page will be used instead
of the default output encoding set for the message catalog. 

=back

=cut

######################################################################

sub Message { 
    my ($self, $msgTag, $encoding) = @_;
    my ($mkey,                          # Message key or literal
        @args);                         # Message key arguments
    my $status;

    if (! defined $msgTag) {
        return "Undefined message tag";
    }

    if (ref($msgTag)) {
        ($mkey, @args) = @{ $msgTag }; 
    } else {
        @args = ($msgTag);
        ($status,@args) = $self->ConvertString(($msgTag),"CORE_CODEPAGE","IO_CODEPAGE"); 
        if ($status) {
            # Unable to complete converting strings while reporting message
            return "ERROR: MsgCat::Message - Unable to convert msgTag from " . 
                   "CORE_CODEPAGE to IO_CODEPAGE\n$status";
        }
        return $args[0];
    }

    # print "MsgCat::Message - processing $mkey\n";

    my %catalog = %{ $self->{CATALOG} };
    my %arguments = %{ $self->{ARGLISTS} };
    my $margs = $arguments{$mkey};   # Expected argument
    my @msgArgs = $margs ? @{$margs} : ();
    my $result;                      # String result 

    if ($self->{SELFTEST}) {
        @args = @msgArgs;

    # Evaluate arguments with "&GetParm('name')" pattern
    } else { 
        my @evaluated_args; 
        for (@args) {
            if ($_ =~ m/^PARM:(\S+)/) {
                push @evaluated_args, &GetParm($1);
            } else {
                push @evaluated_args, $_;
            }
        }
        @args = @evaluated_args;
    } 

    # Check that message key is valid
    if ($catalog{$mkey} eq "") {
        $result = "Message error: Unknown message $mkey\n";

    } elsif (uc(&GetParm("CLEARCASE_ENABLE_ATE_MESSAGES")) eq "TRUE" ) {
        # L10N: Assume that message arguments are from CORE_CODEPAGE
        # standard and that printf (STDOUT) requires local code page
        ($status,@args) = $self->ConvertString(@args,"CORE_CODEPAGE","IO_CODEPAGE"); 
        if ($status) {
            # Unable to complete converting strings while reporting message
            return "ERROR: MsgCat::Message - Unable to convert message " . 
                   "arguments from CORE_CODEPAGE to IO_CODEPAGE\n$status";
        }
        $result = printf "TAG_ID: $mkey: END_ID\n" . 
                         "TAG_VALUES:%s:END_VALUES\n",
                         join("!!",@args);

    } else {
        # L10N: Decide catalog codepage and desired output code page
        my $cp = $self->{MSGCAT_CODEPAGE};

        my $outputcp = ($encoding && $encoding ne "")
                        ? $encoding : "IO_CODEPAGE";

        # L10N: Assume catalog_msg is in local catalog codepage
        # and that arguments are the core codepage and need to 
        # both be converted to destination codepage 
        # then be formatted together

        my $messageFormat;
        ($status,$messageFormat) = 
            $self->ConvertString($catalog{$mkey},$cp,$outputcp);
        if ($status) {
            # Unable to complete converting strings while reporting message
            return "ERROR: MsgCat::Message - Unable to convert catalog " . 
                   "message from $cp to $outputcp\n$status";
        }
        if (@args) { 
            ($status, @args) = $self->ConvertString(\@args,"CORE_CODEPAGE",$outputcp);
            if ($status) {
                # Unable to complete converting strings while reporting message
                return "ERROR: MsgCat::Message - Unable to convert " . 
                       "arguments from CORE_CODEPAGE to $outputcp\n$status";
            }
        }

        # Check that correct number of arguments were provided 
        # Problems with mismatches should be fixed during development
        # but fudge missing arguments for end user if found during use

        if ($#args != $#msgArgs) {
            # Disable for delivery
            # my $n = $#msgArgs + 1;
            # $result = sprintf "Message error: " . 
            #    "$mkey message requires $n arguments \n" . 
            #    "   Arguments expected: @msgArgs\n" . 
            #    "   Arguments provided: @args \n";

            my $start = $#args;
            my $i;
            for ($i = $start + 1; $i <= $#msgArgs; $i++) {
                push @args, $msgArgs[$i];
            }
            $result = sprintf $messageFormat, @args;

        # Print with arguments if needed
        } elsif ($#msgArgs > -1) {
            $result = sprintf $messageFormat, @args;

        # Print with no arguments 
        } else {
            $result = $messageFormat;
        }     
    }
    return $result;
} 

######################################################################
#::method SelfTest

=over 6

=item C<SelfTest(@args)>

This method runs a self test on the message catalog class or subclass.
It will display all messages with their argument names formatted
as if they were argument values to show how the final message will 
lock. The final appearance will vary depending on the length of the
final arguments. 

=back

=cut

######################################################################

sub SelfTest {
    my ($self, @args) = @_;
    my $limit = ($args[0] ? shift : 0);
    my($key, $count, @argsList);
    my %arguments = %{ $self->{ARGLISTS} };

    $self->{SELFTEST} = 1;
    foreach $key (sort keys %arguments) {
        print "########################################\n";
        @argsList = @{ $arguments{$key} };
        $count = $#argsList + 1;
        print "Message $key \narguments: $count (" . 
            join(",",@argsList) . ")\n\n",
            ( $limit ? substr($self->Message([ $key ]),0,$limit) :
                       $self->Message([ $key ])),
            "\n";
    }
    $self->{SELFTEST} = 0;
    return 0;
}

######################################################################
#::method YesNo

=over 6

=item C<YesNo(answer)>

This method compares a given answer to the local values for "yes"
and "no" and returns 0 for neither, 1 for "yes", and 2 for "no".
Local message catalogs can override this method to provide more flexible
matches handling local conventions like case and partial answers,
i.e. "Y" for "yes".

This method tries to find local values for "yes" and "no" by checking
its contents for two tags, IO_P_YES and IO_P_NO. If neither is found, 
returns 0.

=back

=cut

######################################################################

sub YesNo {
    my ($self, $answer) = @_;
    my $result = 0;
    my $yes = $self->Message([ "IO_P_YES" ]);
    my $no = $self->Message([ "IO_P_NO" ]);
    if (defined $yes and defined $no) {
        if ($answer eq $yes) {
            $result = 1;
        } elsif ($answer eq $no) {
            $result = 2;
        }
    }
    return $result;
}

######################################################################
#::method XLIFFwrite

=over 6

=item C<XLIFFwrite(filename)>

This method outputs the current message catalog as an .XLIFF file useful
for translation purposes. A .WC file is also outputted to faciliate word 
counting.

Outputs XLIFF file in: C:\Program Files\Rational\ClearCase\lib\CQCC

Test code: 

$msgCat = new CQCC::MsgCat_English()
$msgCat->XLIFFwrite(filename)

=back

=cut

######################################################################

sub XLIFFwrite {

    my ($self, $filename) = @_;
    my $srcFile = 
        "M:\\bugbert_ccase\\bin\\trigger_cqcc\\CQCC\\MsgCat_English.pm";

    my $dstLang = "EN";
    my $result = 0;
    ## @lines and @wcLines are the otput
    my ($key, $count, @argsList, $arg, @lines, @wcLines);
    my %arguments = %{ $self->{ARGLISTS} };

    push @lines, 
       '<?xml version=\'1.\' encoding=\'utf-8\' ?>',
       '<!DOCTYPE xliff PUBLIC "-//XLIFF//DTD XLIFF//EN"',
       '    "http://www.oasis-open.org/committees/xliff/documents/xliff.dtd" >',
       "<xliff version='1.1'>",
       '<file source-language="EN" datatype="plaintext" ' . 
           "original=\"$srcFile\"",
       " target-language=\"$dstLang\">",
       '    <header></header>',
       '     <body>';

    my $key_counter = 0;
    foreach $key (sort keys %arguments) {
        ## get a list of arguments for each key (trans-unit id)

        $key_counter++;
        @argsList = ();
        for $arg (@{ $arguments{$key} }) {
            if ($arg =~ /\<(\S*)\>/) {
                $arg = $1;
            }
            elsif ($arg =~ /\<(\S+)[\s](\S+)\>/) {    ## some tags are spaced by " " instead of "_"
                $arg = $1 . "_" . $2;
            }
            push @argsList, $arg;
        }        
        $count = $#argsList + 1;
        push @lines, 
            "      <trans-unit id=\"$key\">";
        if (@argsList) {
            push @lines,
                "       <ARGS>" . join(",",@argsList) . "</ARGS>";
        }
        my $message = $self->Message([ $key ]); 

        $message =~ s/</&lt;/g;                    ## escape illegal XML characters
        $message =~ s/>/&gt;/g;

        ## Mark off non-translatable arguments
        my @sortedList = sort { length($b) <=> length($a) } @argsList;
        my $tmp_count = 0;

        ## Sort by argument length, biggest to smallest
        ## Mark off the smallest arguments first, to prevent unexpected overwritting
        foreach $arg (@sortedList) {
            my $symb = "tag!!$tmp_count";
            $message =~ s/$arg/$symb/e;
            $tmp_count++;
        }

        foreach $arg (reverse @sortedList) {
            $tmp_count--;
            my $symb = "tag!!$tmp_count";
            my $replace = '<mrk mtype = \'symbol\'>' . $arg . '</mrk>';
            $message =~ s/$symb/$replace/e;
        }

        push @lines, 
            "       <source xml:lang=\"en\">$message</source>",
            "       <target></target>",
            "      </trans-unit>";
        push @wcLines, $message;
    }
    push @lines,
        "     </body>",
        " </file>",
        "</xliff>";

    my $sep = ($Config::Config{osname} eq "MSWin32") ? "\\" : "/";
    # create output path to atria/lib/CQCC directory

    my $output_path = CQCC::IOBase::IOPathname("ATRIA", ".." . $sep . "lib" . $sep . "CQCC");

    $output_path =~ s/\"//g;

    chdir ($output_path);

    if ( open(OUTFILE, ">$filename" . ".xliff") ) {
        print OUTFILE join("\n",@lines);
        close OUTFILE;

        open (OUTFILE, ">$filename" . ".wc");
        print OUTFILE join("\n",@wcLines);
        close OUTFILE;

        $result = 1;
    }

    return $result;
}

######################################################################
#::Tail
######################################################################

1;
